#!/usr/bin/perl
#   Script Name: Nevil the Pointless Non-Forking Server
#        Author: Richard Terry
#       Website: http://radiac.net/perl
#       Version: 1.0.0
# Latest change: None (release)

#############################
### Initialise the script ###

#####################
## Include modules ##
use strict;
use Socket; # For connections
use IO::Socket; # Also for connections
use IO::Select; # A hell of a lot easier than using the four argument 'select'
use POSIX; # For unblocking and stuff
use Fcntl; # For unblocking
use Class::Struct; # For client struct
use Tie::RefHash; # For ordering the hash


######################
## Set up variables ##
my ($pid, $server_socket);
my (%cmdlist, %config, %clients);
tie %clients, 'Tie::RefHash';
&load_cmdlist; # Load commands
my $fconfig = "nevil.conf"; # Location of configuration file
&struct_client; # Set up client struct (could be long)


#####################
## Predeclare subs ##
sub title;
sub tellu;
sub broadcast;
sub out;
sub prompt;

sub read_file;
sub plural;
sub ordinal;
sub pad;

sub AUTOLOAD { no strict; my ($command, $client_socket) = ($AUTOLOAD, $_[0]); $command =~ s/.*::command_//; tellu $client_socket, " Cannot match command '$command'"; }


#########################
## Read in config file ##
open(FILE,"$fconfig") || &error("Can't open the file '$fconfig'.<BR>Reason: $!\n");
my @lines=<FILE>;
close(FILE);
foreach my $line (@lines) {
	$line =~ s/\n|\r//;
	if (($line !~ /^\#/) && ($line =~ /=/)) {
		my ($var,$val) = split(/=/,$line,2);
		$config{$var} = $val;
	}
}


#########################
## Deal with arguments ##
foreach my $arg (@ARGV) {
	if ($arg eq "-f" || $arg eq "--foreground") {
		$config{'daemon'} = "";
	} elsif ($arg eq "-b" || $arg eq "--background") {
		$config{'daemon'} = "yes";
	} elsif (($arg eq "-r" || $arg eq "--restart") || ($arg eq "-x" || $arg eq "--exit")) {
		if (not -e $config{'pidfile'}) {
			die "Cannot exit, no pidfile";
		} else {
			$pid = (read_file($config{'pidfile'}))[0];
			if ($arg eq "-x" || $arg eq "--exit") {
				kill TERM => $pid;
				print "SIGTERM sent.\n";
			} else {
				kill HUP => $pid;
				print "SIGHUP sent.\n";
			}
			exit;
		}
	}
}

# We're going to try and run... check we're not running already
die "Nevil already seems to be running as ".(read_file($config{'pidfile'}))[0].". Terminating...\n" if (-e $config{'pidfile'});


##################
## Load plugins ##
&load_plugins($config{'plugins dir'}) if (exists($config{'plugins dir'}));


########################
## Process management ##
# Ensure SIGHUP is unblocked
sigprocmask(SIG_UNBLOCK, POSIX::SigSet->new(SIGHUP)) or die "Can't unblock SIGHUP: $!\n";
fork and exit if ($config{'daemon'} eq "yes"); # If we want to run in the background, now's the time... but we don't want it to fork on a reload
{ no strict;
	# Program name (ps -aux)
	$PROGRAM_NAME = "Nevil";
}
$SIG{HUP} = \&sighup; # We want to be able to restart
$SIG{INT} = \&sigterm; # We want to control the termination
$SIG{TERM} = \&sigterm; # We want to control the termination
$pid = $$;


######################################
## Open socket ready for connection ##
$server_socket = IO::Socket::INET->new(LocalPort => $config{'port'}, Listen => SOMAXCONN, Reuse => 1) or die "Can't make server socket: $@\n";
&nonblock($server_socket);

# We are running as server...
if ($config{'pidfile'} ne "") {
	open (PID,">$config{'pidfile'}") || die "Couldn't open log file - $!\n";
	print PID $pid;
	close PID;
}

# If we're going to be logging, start logging now
if ($config{'outpipe'} ne "") {
	open (FILE,">$config{'outpipe'}") || die "Couldn't open log file - $!\n";
	select(FILE); # Select $client_socket channel
	$| = 1; # Force a flush after every write to the channel
}

# Send output to say we're ready...
out "^gInitialised^n as^y $$";
out "^gSocket open^n on port $config{'port'}";


#########################
### Main loop section ###
# Accept and process connections
my $select = IO::Select->new($server_socket);

while (1) {
	foreach my $client_socket ($select->can_read(1)) {
		my ($in);
		
		if ($client_socket == $server_socket) {
			# Accept new connection
			$client_socket = $server_socket->accept();
			$select->add($client_socket);
			&nonblock($client_socket);
			
			# Add them to the list
			my $client = new client;
			$client->socket($client_socket);
			$clients{$client_socket} = \$client;
			
			# Find out about our new connection
			$client->ip($client_socket->sockhost);
			#$port = $client_address->sockport; # In case you want it :)
			
			# Say something to them
			out "^gAccepted^n - connection from ^p".$client->ip;
			tellu $client_socket, title("Nevil the Pointless Non-Forking Server");
			tellu $client_socket, "^wHello and welcome to ^yNevil^w the ^pPointless ^cNon-Forking ^gServer^w!";
			tellu $client_socket, "Type 'commands' to list available commands.\n^L";
			
		} else {
			# Existing connection, read the port
			my $rec = $client_socket->recv($in, POSIX::BUFSIZ, 0);
			
			unless ($rec && length $in) {
				# The client has dropped, lets kill them
				&kill_client($client_socket);
				next;
			}
			
			# Client is ok
			my $client = $clients{$client_socket};
			$in =~ s/\xFF..//g; # Strip out control stuff, can't be bothered with that :p
			my $tinbuffer = $$client->inbuffer . $in;
			my @inlines = ();
			while ($tinbuffer =~ s/(.*)\r\n//) {
				push(@inlines, $1);
			}
			$$client->inbuffer($tinbuffer);
			
			if (scalar(@inlines) > 0) {
				foreach my $line (@inlines) {
					# Please note: a ' at the start of the line will cause an error when it tries to call the subroutine
					$line =~ s/^'//g;
					my ($command, @vars) = split(/ /, $line);
					{ no strict;
						&{"command_$command"}($client_socket, @vars);
					}
				}
				$$client->inbuffer(undef);
			}
		}
	}
		
	# Anything to send?
	foreach my $client_socket ($select->can_write(1)) {
		my $client = $clients{$client_socket};
		next unless ($$client->outbuffer);
		
		# Let's send it
		my ($out, $status) = &transmit($client_socket, $$client->outbuffer);
		next if ($status); # If we killed the client
		$$client->outbuffer($out);
	}
}
out "^rTerminating^n - No more connections";
close $server_socket;
close FILE if ($config{'outpipe'} ne "");
exit;


###################
### Subroutines ###

###################
## Core routines ##
# (Aka 'Stuff that does stuff')

sub read_file {
	open(FILE,"<$_[0]") || &error("Can't read from the file '$_[0]'.<BR>Reason: $!\n");
	my @lines=<FILE>;
	close(FILE);
	return @lines;
}

sub out {
	my $out = "^c$$^n> $_[0]";
	if ($config{'logcolour'} eq "yes") {
		$out = &trans($out);
	} else {
		$out =~ s/\^[drgybpcwn]//g;
		$out = &trans($out);
	}
	print "$out\n";
}

sub tellu {
	# Tell the specified user something
	# Usage: tellu $client_socket, "whatever";
	my ($client_socket,$out) = @_;
	$out = &trans($out);
	$out =~ s/\n/\r\n/g;
	die "no such client: $client_socket" unless (exists($clients{$client_socket}));
	my $client = $clients{$client_socket};
	$$client->outbuffer($$client->outbuffer . "$out\r\n");
}

sub trans {
	my $out = $_[0];
	
	# Dynatext
	$out =~ s/(^|[^\^])\^L/$1^y$config{'line'}\e[0m/g; # Line
	
	# Colours
	my $col = &col($out);
	$out = $col . "\e[0m" if ($col ne $out);
	
	$out =~ s/\^\^/\^/g;
	
	return $out;
}

sub col {
	# Codes from http://www.linuxgazette.com/issue01to08/articles.html
	my $c = $_[0];
	$c =~ s/(^|[^\^])\^d/$1\e[1;30m/g; # black
	$c =~ s/(^|[^\^])\^r/$1\e[1;31m/g; # red
	$c =~ s/(^|[^\^])\^g/$1\e[1;32m/g; # green
	$c =~ s/(^|[^\^])\^y/$1\e[1;33m/g; # yellow
	$c =~ s/(^|[^\^])\^b/$1\e[1;34m/g; # blue
	$c =~ s/(^|[^\^])\^p/$1\e[1;35m/g; # purple
	$c =~ s/(^|[^\^])\^c/$1\e[1;36m/g; # cyan
	$c =~ s/(^|[^\^])\^w/$1\e[1;37m/g; # white
	$c =~ s/(^|[^\^])\^n/$1\e[0m/g; # normal
	return $c;
}

sub title {
	my ($in) = @_;
	return $in if (length($in) > 74);
	$in = " $in ";
	my $out = $config{'line'};
	substr($out, (length($out)/2)-(length($in)/2), length($in)) = $in;
	$out =~ s/ ([\=\-\*]+)$/ \^y$1/ if ($in =~ /\^/);
	return "\n\r^y$out";
}

sub broadcast {
	# Sub to broadcast to everyone, with optional exclude
	my ($exclude, $out) = @_;
	($exclude, $out) = ($out, $exclude) unless ($out);
	# Tell the users what's happening...
	foreach my $client_socket (keys %clients) {
		next if ($client_socket == $exclude);
		my $client = $clients{$client_socket};
		$$client->outbuffer($$client->outbuffer . &trans($out) . "\r\n");
	}
}

sub transmit {
	my ($client_socket, $out) = @_;
	my $sen = $client_socket->send($out, 0);
	
	# See what's happened
	unless (defined $sen) {
		out "^pFailed to write^n - was told I could...";
		next;
	}
	
	# Well... something happened...
	if ($sen == length($out) || $! == POSIX::EWOULDBLOCK) {
		# Send was OK. Might not have sent it all because we might have run out of space.
		substr($out, 0, $sen) = '';
		return $out;
	} else {
		# Send failed for a different reason
		&kill_client;
		return (undef, 1);
	}
}

sub flush_transmit {
	my ($client_socket, $message) = @_;
	my $client = $clients{$client_socket};
	my $sent = 0;
	$$client->outbuffer($$client->outbuffer . &trans($message) . "\r\n");
	while ($sent == 0) {
		my ($out, $status) = &transmit($client_socket, $$client->outbuffer);
		if ($status) {
			# If we killed the client
			$sent = undef;
		} else {
			$sent = 1 if (length($out) == 0);
			$$client->outbuffer($out);
		}
	}
	return $sent;
}

sub kill_client {
	# Kill $client_socket
	my ($client_socket, $message) = @_;
	my $client = $clients{$client_socket};
	
	$select->remove($client_socket);
	
	delete $clients{$client_socket};
	
	$message = "^rDisconnecting^n - ".$$client->ip." disconnected" unless $message;
	out "$message";
	$client_socket->close;
}

sub nonblock {
	my $flags = fcntl($_[0], F_GETFL, 0) or die "Can't get flags for socket: $!\n";
	fcntl($_[0], F_SETFL, $flags | O_NONBLOCK) or die "Can't make socket nonblocking: $!\n";
}

sub sighup {
	$SIG{HUP} = \&sighup; # We want to be able to reload
	
	# Tell the log what's happening...
	out "^pRestarting^n - ^pSIGHUP^n recieved";
	
	# Tell the users what's happening...
	foreach my $client_socket (keys %clients) {
		my $sent = &flush_transmit($client_socket, "^rRestarting^n - ^rSIGHUP^n recieved");
		next unless $sent;
		my $client = $clients{$client_socket};
		&kill_client($client_socket, "^pDisconnecting ".$$client->ip."^n - restarting");
	}
	
	# Close the port
	$server_socket->close;
	
	# Find out the file address and run ourselves again
	my $fself .= $ENV{'_'};
	$fself =~ s/^\.//;
	exec "$ENV{'PWD'}$fself", "-f"; # Exec this as 'foreground' - we don't want it to refork
	exit; # Shouldn't be needed, but in case the exec failed...
}

sub sigterm {
	out "^rTerminating^n - ^rSIGTERM^n recieved";
	
	# Tell the users what's happening, and disconnect them
	foreach my $client_socket (keys %clients) {
		my $sent = &flush_transmit($client_socket, "^rTerminating^n - ^rSIGTERM^n recieved");
		next unless $sent;
		my $client = $clients{$client_socket};
		&kill_client($client_socket, "^rDisconnecting ".$$client->ip."^n - terminating");
	}

	# Close the port
	$server_socket->close;
	
	unlink($config{'pidfile'}) unless ($config{'pidfile'} eq "");
	
	exit;
}

sub error {
	out "^rTerminating^n - ^rFatal Error:^n $_[0]";
	
	# Tell the users what's happening, and disconnect them
	foreach my $client_socket (keys %clients) {
		my $sent = &flush_transmit($client_socket, "^rTerminating^n - ^rFatal Error^n");
		next unless $sent;
		my $client = $clients{$client_socket};
		&kill_client($client_socket, "^pDisconnecting ".$$client->ip."^n - terminating");
	}
	
	# Close the port
	$server_socket->close;
	
	unlink($config{'pidfile'}) unless ($config{'pidfile'} eq "");
	
	exit;
}


#########################
## Peripheral Routines ##
# (Aka 'The stuff that makes it useful')

sub struct_client {
	# Set up client struct
	# Vars in the format: varname vartype
	struct client => { reverse qw{
		$ socket
		$ inbuffer
		$ outbuffer
		$ ip
	}};
}

sub load_cmdlist {
	# Hash of commands and descriptions for use with command_commands and command_describe
	%cmdlist = (
	"commands", "List all commands",
	"colours", "Shows the different colours available",
	"describe", "Describes a command",
	"quit", "Disconnects you from the program"
	);
}

sub command_commands {
	my ($client_socket, @vars) = @_;
	my @commands;
	foreach my $key (sort keys %cmdlist) {
		push (@commands, $key);
	}
	tellu $client_socket, title("Commands");
	tellu $client_socket, join(", ", @commands);
	tellu $client_socket, "^L";
}

sub command_colours {
	my ($client_socket, @vars) = @_;
	tellu $client_socket, title("Available Colours");
	my @colours = qw(dark red green yellow blue purple cyan white normal);
	foreach my $colour (@colours) {
		$colour =~ /^(.)/;
		tellu $client_socket, "$colour (^^$1): ^$1hello";
	}
	tellu $client_socket, "And don't forget, ^^L will give you a line.\n^L";
}

sub command_describe {
	my ($client_socket, @vars) = @_;
	my $command = shift(@vars);
	unless (exists($cmdlist{$command})) {
		tellu $client_socket, " Cannot match command '$command'";
		return;
	}
	tellu $client_socket, title("Description of '$command'");
	tellu $client_socket, "$cmdlist{$command}\n^L";
}

sub command_secret {
	my ($client_socket, @vars) = @_;
	tellu $client_socket, "Woo! You found the secret command!";
}

sub command_quit {
	my ($client_socket, @vars) = @_;
	tellu $client_socket, title("Goodbye!");
	&flush_transmit($client_socket, "Bye... come again soon!\n^L\n");
	&kill_client($client_socket, "Client quit");
}
